#!/opt/edge/venv/bin/python3

import sys
import os
import getopt
import math
import time
import struct
from pymodbus.client.serial import ModbusSerialClient

endpoint = '/dev/ttyS1'
slave_addr = 0x31  # 0x31 = 49

verbose = False


def unpack_wiegand_syms(result, number, base):
    div_base, mod_base = divmod(number, base)
    if div_base < base:
        result.extend([mod_base, div_base])

    else:
        result.append(mod_base)
        unpack_wiegand_syms(result, div_base, base)

    return result


def _xor_bits(data, max_bit, min_bit):
    assert type(data) == int
    assert 0 <= min_bit <= max_bit and max_bit >= 0

    data = data >> min_bit
    crc = data & 0x01
    for i in range(min_bit, max_bit + 1):
        data = data >> 1
        crc = crc ^ (data & 0x01)
    return crc


def _mk_wiegand_package(payload_data, payload_size):
    """
    :param payload_data: payload24, payload32, ...
    :param payload_size: 24, 32, ...
    :return: package26, package34, ...
    """
    assert type(payload_data) == int
    assert payload_size >= 24

    xor_hi = _xor_bits(payload_data, payload_size - 1, payload_size / 2)
    xor_lo = _xor_bits(payload_data, payload_size / 2 - 1, 0)

    parity_hi = xor_hi << (payload_size + 1)
    parity_lo = (~xor_lo) & 0x01

    return parity_hi | (payload_data << 1) | parity_lo


def bmask(n):
    if n < 1:
        return 0

    r = 0x01
    n -= 1
    while n > 0:
        r = r << 1 | 0x01
        n -= 1
    return r


def _mk_payload(plate, max_payload_size, charset, min_char_size):
    """
    :param plate: plate_text
    :type plate: str

    :param max_payload_size: max-bits
    :type max_payload_size: int

    :type charset: dict(list)
    :type min_char_size: int
    """
    payload_data = 0
    power = 1
    for c in plate.decode('utf-8').upper():
        sym_ucode = ord(c)
        sym_lohi = charset.get(sym_ucode, [0xFF, 0xFF])
        if sym_lohi and sym_lohi[0] < 0xFF:
            for i in range(2):
                if sym_lohi[i] < 0xFF:
                    payload_data += sym_lohi[i] * power
                    power *= min_char_size
                    if verbose:
                        print(("%s=%s: %d: %d ---> %d" % (c, sym_ucode, i, sym_lohi[i], payload_data)))
        else:
            print("fail: all symbols should be from the Unicode Basic Plane")
            dd = sym_ucode
            while dd > min_char_size * min_char_size:
                b = dd % min_char_size
                payload_data += b * power
                power *= min_char_size
                dd /= min_char_size

    payload_size = math.ceil(math.log(payload_data) / math.log(2))
    while True:
        m = math.ceil(math.log(payload_data) / math.log(2))
        if m > max_payload_size:
            k = m % max_payload_size if m < 2 * max_payload_size else max_payload_size
            payload_data = int((payload_data ^ ((payload_data & bmask(int(k))) << int(m - k))) >> int(k))
        else:
            break

    return payload_data, payload_size


def _load_charset(schema_path):
    """
    :type schema_path: str
    :return alphabet table and alphabet size
    :rtype: (dict(list), int)
    """
    with open(schema_path, 'rb') as schema_file:
        be_data = schema_file.read()  # big-endian

    """
    $ hexdump -e '4/1 "%02X " /1 " | %02X" /1 " %02X\n" ' ../../share/misc/def-charset.bin | grep "00 00 04 27"
    00 00 04 27 | 03 08
    
    $ hexdump -e '4/1 "%02X " /1 " | %02X" /1 " %02X\n" ' ../../share/misc/def-charset.bin | grep "00 00 00 42"
    00 00 00 42 | 02 FF
    """
    sym_ucode_to_lohi_list = dict()

    offset = 0
    unique_1_sym = set()
    while offset < len(be_data):
        # unpack big-endian unsigned int
        sym_ucode_list = struct.unpack_from('>I', be_data, offset)
        offset += 4

        # unpack unsigned char, then unsigned char
        sym_lohi_list = struct.unpack_from('BB', be_data, offset)
        offset += 2

        if sym_lohi_list[1] == 0xFF:
            unique_1_sym.add(sym_lohi_list[0])

        sym_ucode_to_lohi_list[sym_ucode_list[0]] = sym_lohi_list

    return sym_ucode_to_lohi_list, len(unique_1_sym)


def usage(msg=None):
    if msg:
        print(("ERROR: " + str(msg)))

    print(
        """
            wiegand.py 
                -h              help
                
                -k              print key string for controller
                
                -w [0,1]        write to wiegand port default=none
                --wiegand=
                
                -b [24,32,40] key length default=24 (wiegand: 26,34,42)      
                --bits=
                
                -p [plate]      plate (utf8) to hash
                --plate=
                
                -f [file]       path to bin file with alphabet
                --binfile=
                
                --csv           use .csv files instead of STDIN and STDOUT
                
                --in [path]     path to input file with numbers, each number in a new line
                                Default to './plates.csv'
                                
                --out [path]    path to output CSV file with hashes
                                Default to './hashes.csv'
        """)
    sys.exit(1)


def trykey(mi, p, max_payload_size, w, alphabet, a_size):
    payload_data, payload_size = _mk_payload(p, max_payload_size, alphabet, a_size)
    package_data = _mk_wiegand_package(payload_data, max_payload_size)

    _wds = int(math.ceil((max_payload_size + 2) / 16.0))
    package_data <<= (_wds * 16 - max_payload_size - 2)
    r = []
    for _w in range(_wds):
        r.insert(0, package_data & 0xffff)
        package_data >>= 16

    with mi as modbus:
        modbus.write_register(w, max_payload_size + 2, unit=slave_addr)
        time.sleep(0.5)
        modbus.write_registers(0 if w == 0 else 0x20, r, unit=slave_addr)


def _mk_bolid_payload(plate, max_payload_size, charset, min_char_size):
    payload_data, payload_size = _mk_payload(plate, max_payload_size, charset, min_char_size)
    return "%014X" % ((payload_data << 8 | 0x01) & bmask(64)), payload_size


def _main():
    options, args = getopt.getopt(
        sys.argv[1:], 'hHkvw:b:p:f:',
        ['help', 'key', 'verbose', 'csv', 'in=', 'out=', 'wiegand=', 'bits=', 'plate=', 'binfile='])

    global verbose
    keyonly = False
    plate = None
    wiegand = None
    max_payload_size = 24
    bin_filepath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'def-charset.bin')
    plates_filepath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'plates.csv')
    hashes_filepath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'hashes.csv')
    use_csv = False

    try:
        for opt, value in options:
            if opt in ['-h', '-H', '--help']:
                usage()
            elif opt in ['-v']:
                verbose = True
            elif opt in ['--csv']:
                use_csv = True
            elif opt in ['--in']:
                plates_filepath = value
            elif opt in ['--out']:
                hashes_filepath = value
            elif opt in ['-k', '--key']:
                keyonly = True
            elif opt in ['-w', '--wiegand']:
                wiegand = int(value)
            elif opt in ['-b', '--bits']:
                max_payload_size = int(value)
            elif opt in ['-p', '--plate']:
                plate = value
            elif opt in ['-f', '--binfile']:
                bin_filepath = value

    except getopt.error as msg:
        usage(msg)

    if not bin_filepath:
        usage("File is mandatory")

    if not plate and not use_csv:
        usage("Plate is mandatory")

    charset, min_char_size = _load_charset(bin_filepath)

    if use_csv:

        print('using [{}] as input file'.format(plates_filepath))
        print('using [{}] as output file\n'.format(hashes_filepath))

        plates_csv = open(plates_filepath, 'r')
        hashes_csv = open(hashes_filepath, 'w')
        for plate in plates_csv:
            plate = plate.strip()
            payload_data, payload_size = _mk_payload(plate, max_payload_size, charset, min_char_size)
            print('hash [{}] to [{:08X}] using wiegand {}'.format(plate, payload_data, max_payload_size + 2))
            hashes_csv.write('{},{:08X}\n'.format(plate, payload_data))

        plates_csv.close()
        hashes_csv.close()

        sys.exit(0)

    if keyonly:
        # payload_data, payload_size = _mk_bolid_payload(plate, max_payload_size, charset, min_char_size)
        payload_data, payload_size = _mk_payload(plate, max_payload_size, charset, min_char_size)

        # payload_data = 0x0026A89E  # payload24:0x0026A89E => package26:0x24D513D
        package_data = _mk_wiegand_package(payload_data, max_payload_size)
        print('payload{}:{:08X}, package{}:{:08X}'.format(
            max_payload_size, payload_data, max_payload_size + 2, package_data))

        if verbose:
            print((int(payload_size)))
        sys.exit(0)

    if wiegand is None:
        usage("Wiegand port is mandatory")

    mi = ModbusSerialClient(method="rtu", port=endpoint, baudrate=19200, bytesize=8, stopbits=1, parity='N')
    mi.timeout = 1

    trykey(mi, plate, max_payload_size, wiegand, charset, min_char_size)
    sys.exit(0)


if __name__ == "__main__":
    _main()
